/*****************************************************************************

LOADER_GF Rev 0.02


IMPORTANT: be sure "Include debug code RST/28 instructions" is NOT checked
on the <Compile> menu!

then <Compile> "Compile To Target"

  Run with a serial utility such as Kermit.
  Configure the serial utility for port connected to RS232, 19200 8N1.



 10/20/01	Initial coding.
 10/24/01	Changed assy parameter for baud & result to HL
 10/25/01	Hard coded flash buffer for extension code
 01/26/02	Test MMU remap after xmem2xmem
 01/29/02	forth now runs with fully mapped ram at $0000
 				No BIOS customization needed.
 04/15/03	Serial port selection via define
 				Better dialog on flash exit
 06/03/03	Changed to FS2 in program Flash for image (was WriteFlash)
 				Must edit BIOS again (for FS2:  XMEM_RESERVE_SIZE = 0x8000)
 05/27/04 	K.I.S.S. - No FS2 (no BIOS edit). Only serial port A.
 07/02/04   Fix assy use of vars must be static (DC8.x changed default)
 				Started versioning file names.
 07/08/04	Rev 0.02: Changed RAM allocation to 56K.

*****************************************************************************/

#define TRUE 1
#define FALSE 0

#define DEBUG_PRINT FALSE		// ENABLE stdio print?


#if DEBUG_PRINT == TRUE
#define PR_DEBUG(string)  printf( string ) ;
#else
#define PR_DEBUG(string)
#endif

#if DEBUG_PRINT == TRUE
#define PR_DEBUG1(string,arg)  printf( string , arg ) ;
#else
#define PR_DEBUG1(string,arg)
#endif






// use xmem for main()!!
//
// it is critical that the in-line
// assembly that reconfigures the MMU and calls FORTH
// be located in xmem (0xE000-0xFFFF) so it is still mapped
// during the flash to ram and back change
//

#memmap xmem


// define buffer sizes for serial port

#define AINBUFSIZE 15
#define AOUTBUFSIZE 63



#define BAUDRATE 19200


/*****************************************************************************/

#ximport "gfr2k.bin" forthbin


// places the length of <filename> ( stored as a long)
// followed by its binary contents at the next available place in xmem flash.
// forthbin is a 'constant MACRO' that resolves to a long (address).

// gforth is about 18k bytes.


/*****************************************************************************/


#define FORTHORG 0x0000

// forth is in assy. org to start at 0x0000 (in RAM).



/*****************************************************************************

	Routines for serial port

*****************************************************************************/

int SerialPuts(char *str)
{
  return(serAputs(str));
}

int SerialwrFree()
{
  return(serAwrFree());
}



#define PR_SERIAL(string)\
      SerialPuts( string );\
      while (SerialwrFree() != AOUTBUFSIZE) ;

#define PR_SERIAL1(string,arg)\
		sprintf(G_buf, string, arg);\
      SerialPuts( G_buf );\
      while (SerialwrFree() != AOUTBUFSIZE) ;




/*****************************************************************************

	Routines for segment read

*****************************************************************************/

int get_dataseg( )
{
#asm nodebug
	ioi ld a,(DATASEG)
	ld l,a
	ld h,0
#endasm
}

int get_stackseg( )
{
#asm nodebug
	ioi ld a,(STACKSEG)
	ld l,a
	ld h,0
#endasm
}

int get_segsize( )
{
#asm nodebug
	ioi ld a,(SEGSIZE)
	ld l,a
	ld h,0
#endasm
}

// (root) data builds down from (watch_code_base - 1)

// keep out of:
// extended memory logical address region E000H to FFFFH





/*****************************************************************************


							MAIN


*****************************************************************************/

// the stdio printf output code is executed during the
// 1st run of the program while in debug (RST28 instructions) mode
// (attached to PC running Dyn-C).

// The gForth binary image is loaded into xmem at compile time.


xmem main()
{
	int i;
	int c;
	int err;
	int addr;	// return value from FORTH

   long baud;
   long divisor;

   unsigned long addr1;
   unsigned long addr2;


   unsigned long length;
   unsigned int len16;


   unsigned char nss;
   unsigned char nds;
   unsigned char oss;
   unsigned char ods;


// these vars **must** be static so they can be referenced by address
// from in-line assembly code.

   static int br_div;

   static unsigned int segsize_val;
   static unsigned int dataseg_val;






// INITIALIZE some vars...


// allocate 56K + 4K of ram.
// we can't use all of this (because we MUST keep out of
// xmem space 0xE000..0xFFFF). the extra 4K alloc is to make room
// for 4K alignment.

	addr1 = xalloc(0xF000);

	if (addr1 == -1)
	  {
	  PR_DEBUG("failed to allocate RAM!\n")
	  exit(-1);
	  }
	else
	  {
     PR_DEBUG1("physical address allocated: %06lx\n",(long)addr1)
	  }


// END OF INITIALIZATION


	xmem2root(&length,forthbin,sizeof(long));		// set length (var in root)
																// for copying image into ram

	if (length > 0xFFFF)
	  {
	  PR_DEBUG("Illegal binary image length!\n")
	  exit(-1);
	  }

	len16 = (int)length;									// not more than 64K by design!
	PR_DEBUG1("length = %04X\n",len16)


	// now get ready to load forth from flash into ram, then boot forth from ram...

   ods=get_dataseg();

	PR_DEBUG1("DATASEG: %02x\n",ods)


   i=get_stackseg();

   PR_DEBUG1("STACKSEG: %02x\n",i)


   oss=get_segsize();

  	PR_DEBUG1("SEGSIZE: %02x\n",oss)
	PR_DEBUG1("physical address: %06lx\n",(long)forthbin)


	// baud rate is a parameter.
	// enables changing baud rate without recompiling forth.

   baud = BAUDRATE;
	// (divisor+1) for 19200 baud is stored in BIOS variable "freq_divider"
	divisor = (long)(freq_divider * 19200.0/(float)baud + 0.5) - 1L;
   // The counter divides by (n+1)

   br_div = (int)divisor;

	PR_DEBUG1("baud divisor: %d\n",br_div)
	PR_DEBUG1("BIOS freq_divider: %d\n",freq_divider)

	// change segsize...keeps
	// vars where they are & makes more ram available...
	// forth can now grow up from 0 to bottom of vars

	// put forth at next 4K boundary..._in_ the 56K+4K of ram allocated earlier
   // we don't check if already on a 4K boundary.
   // if ((addr1 & 0xFFF) == 0) /* already aligned */

	addr2=(((addr1 >> 12) + 1) * 0x1000)+FORTHORG;

	PR_DEBUG1("run code at pa: %06lx\n",(long)addr2)

	nds=(unsigned char)(addr2 >> 12);		// calculate new DATASEG

	// NOW LOAD 4K aligned RAM with forth from FLASH
	xmem2xmem(addr2,forthbin+sizeof(long),len16);

	PR_DEBUG("starting forth...\n")
	PR_DEBUG("dummy line\n")
	PR_DEBUG("dummy line2\n")		// so we can see the 2nd previous line before d-c times out

	// structure old and new values for SEGSIZE & DATASEG
	// for in-line assy use...

	segsize_val=((oss & 0xF0) << 8)+ oss;		// (new * 0x100 + original)
	dataseg_val=(nds << 8)+ ods;



// boot forth image from ram...

#asm nodebug

   // stop ints & wdt before changing mem map!
	ld a,0x53
	ioi ld (WDTTR),a
	ld a,0x54
	ioi ld (WDTTR),a			; NO WATCHDOG!

	ioi ld a,(GCSR)
	and	0xFC					; no periodic ints!!!
	ioi	ld (GCSR),a


	ld	hl,(br_div)				; set up parameters

	ld bc,(segsize_val)
	ld de,(dataseg_val)

   // new MMU mapping
	ld A,b
	ioi	ld (SEGSIZE),A
	ld A,d
	ioi	ld (DATASEG),A

   // SP adjust to fully use 56K RAM

   ld (0xDFFE),SP	; stick SP in a safe place
   ld SP, 0xDFFC	; this is where the return address will go

   // START FORTH!

	call	FORTHORG				;with baudrate parameter in HL

   ld SP,(0xDFFE)	; restore original SP

   // restore MMU mapping
	ld A,e
	ioi	ld (DATASEG),A
	ld A,c
	ioi	ld (SEGSIZE),A

	ld (addr),hl				; get return value

#endasm



// forth exit returns to here

// return code:
//				0		no change to image (bye)
//				-1		re-install original image (forget)
//				addr	flash the new image (flash)


   serAopen(baud);

	PR_SERIAL("\r\nDone...\r\n")

   serAclose();

} // end main


